import json
import os


def _add_cls(install, app):
    if install.cls:
        os.makedirs('./app/extension/' + app, exist_ok=True)
        for item in install.cls:
            with open(f'./app/extension/{app}/{item[0]}.py', 'a') as file:
                f = ''
                f += f'from {item[1]} import {item[2]}\n\n'
                f += f'class {item[3]}({item[2]}):\n'
                f += '    pass\n'
                file.write(f)


def _add_imprt(install, app):
    if install.imprt:
        os.makedirs('./app/extension/' + app, exist_ok=True)
        for item in install.imprt:
            with open(f'./app/extension/{app}/{item[0]}.py', 'a') as file:
                f = f'from {item[1]} import {item[2]}\n'
                file.write(f)


def _add_route(install, app):
    if install.route:
        os.makedirs('./app/extension/' + app, exist_ok=True)
        with open(f'./app/extension/{app}/route.py', 'w') as file:
            f = 'import ' + install.route + '\n'
            file.write(f)

        with open('./routes/index.py', 'r') as file:
            all = file.read()
            routes = 'import ' + f'app.extension.{app.replace("/",".")}.route' + '\n'
            all = routes + all
            with open('./routes/index.py', 'w') as b:
                b.write(all)


def _add_admin_templates(install):
    with open('./frontend/src/router/admin.js', 'r') as file:
        all = file.read()
        for item in install.admin_templates:
            all = all.replace('/* {{place new import}} */',
                              '/* {{place new import}} */\nconst ' + item.get('name',
                                                                              'test') + ' = () => import("' + item.get(
                                  'file', 'test') + '");')
            all = all.replace('/* {{place new Route}} */',
                              '/* {{place new Route}} */\n' + item.get('path', '{}') + ',')

        with open('./frontend/src/router/admin.js', 'w') as b:
            b.write(all)


def _add_index_templates(install):
    with open('./frontend/src/router/index.js', 'r') as file:
        all = file.read()
        for item in install.index_templates:
            all = all.replace('/* {{place new import}} */',
                              '/* {{place new import}} */\nconst ' + item.get('name',
                                                                              'test') + ' = () => import("' + item.get(
                                  'file', 'test') + '");')
            all = all.replace('/* {{place new Route home}} */',
                              '/* {{place new Route home}} */\n' + item.get('path', '') + ',')
        with open('./frontend/src/router/index.js', 'w') as b:
            b.write(all)


def _add_user_templates(install):
    with open('./frontend/src/router/index.js', 'r') as file:
        all = file.read()
        for item in install.user_templates:
            all = all.replace('/* {{place new import}} */',
                              '/* {{place new import}} */\nconst ' + item.get('name',
                                                                              'test') + ' = () => import("' + item.get(
                                  'file', 'test') + '");')
            all = all.replace('/* {{place new Route user}} */',
                              '/* {{place new Route user}} */\n' + item.get('path', '') + ',')
        with open('./frontend/src/router/index.js', 'w') as b:
            b.write(all)


def _add_db(install, app):
    if install.db:
        for m, db in install.db.items():
            m = m.split('.')
            name = m[-1]
            m.pop(-1)
            path = '.'.join(m)
            model = __import__(path, fromlist=[''])
            with open(f'extension/{app}/{db}', 'r') as db:
                d = json.loads(db.read())
                model = getattr(model, name)('app')
                for item in d:
                    model.create(item)



import os
from datetime import datetime, timedelta
import pickle
from hashlib import sha3_256

from renus.core.log import Log
from renus.util.helper import encode64

from renus.core.config import Config

try:
    import redis
    rds = redis.Redis()
except ImportError:
    rds=None

class Cache:
    depth = 1
    typ=Config('app').get('cacheDriver','file')
    if typ=='redis':
        assert rds is not None, "'redis' must be installed for Cache"
        rd=rds
    else:
        rd=None
    def __init__(self, prefix: str = '', use_hash=False) -> None:
        self._prefix = prefix
        self._use_hash = use_hash

    def put(self, key: str, value, expire: int = 60):
        """
        add value by key to cache system
        :param key: key for value
        :param value: value that save by key
        :param expire: seconds to expire from now
        """
        if self.rd:
            v=pickle.dumps(value,pickle.HIGHEST_PROTOCOL)
            self.rd.set(self._prefix+key,v,expire)
            return
        expire = datetime.utcnow() + timedelta(seconds=expire)
        path = self._build_name(key, self.depth)
        self._create_if_not_exist(path, value, expire)

    def get(self, key: str, default=None):
        if self.rd:
            r=self.rd.get(self._prefix+key)
            return default if r is None else pickle.loads(r)
        path = self._build_name(key, self.depth)
        return self._read_key(path, default)['value']

    def expire(self, key, default=None):
        if self.rd:
            return max(-1,self.rd.ttl(self._prefix+key))
        path = self._build_name(key, self.depth)
        t= self._read_key(path, default)['expire']
        if t==-1:
            return -1
        return max(-1,int(t.timestamp()-datetime.utcnow().timestamp()))

    def delete(self, key):
        if self.rd:
            return self.rd.delete(self._prefix+key)

        path = self._build_name(key, self.depth)
        return self._delete_file("storage/cache" + path)

    def delete_expired(self):
        filename = './storage/cache/'
        n = 0
        for dirname, subdirs, files in os.walk(filename):
            for file in files:
                p=f'{dirname}/{file}'.replace(filename,'')
                r=self._read_key(p)
                if r['expire']==-1:
                    t= self._delete_file("storage/cache/"+p)
                    if t:
                        n+=1

        Log().info(f'delete_expired success: {n} files')

    def _delete_file(self, path, n=0):
        path = path.strip(' \n')
        try:
            if os.path.exists(path):
                os.remove(path)
        except Exception as exc:
            if n < 3:
                return self._delete_file(path, n + 1)
            Log().info(f'cache delete_file: {exc}')
        return True

    def _check_pass(self, line):
        res = line.split('<=>', 1)
        if len(res) < 2:
            return True
        expired, path = res
        if (datetime.strptime(expired, '%Y-%m-%d %H:%M:%S').timestamp() < datetime.utcnow().timestamp()):
            self._delete_file('storage/cache' + path)
            return False
        return True


    def _create_if_not_exist(self, path, value, expire, n=0):
        filename = 'storage/cache' + path
        try:
            os.makedirs(os.path.dirname(filename), exist_ok=True)
            with open(filename, 'wb') as output:
                pickle.dump(expire, output, pickle.HIGHEST_PROTOCOL)
                pickle.dump(value, output, pickle.HIGHEST_PROTOCOL)
        except:
            if n < 3:
                self._create_if_not_exist(path, value, expire, n + 1)

    def _read_key(self, path, default=None, n=0):
        filename = 'storage/cache' + path
        if not os.path.exists(filename):
            return {'value': default, 'expire': -1}

        try:
            with open(filename, 'rb') as input:
                expire = pickle.load(input)
                value = pickle.load(input)
        except Exception:
            if n < 3:
                return self._read_key(path, default, n + 1)
            else:
                return {'value': default, 'expire': -1}

        if expire.timestamp() < datetime.utcnow().timestamp():
            self._delete_file(filename)
            return {'value': default, 'expire': -1}
        return {'value': value, 'expire': expire}

    def _build_name(self, key: str, depth):
        h_key = hash_key(key, self._use_hash)
        res = self._prefix.strip('/')
        d = 1
        s = 0
        e = 2
        while d < depth + 1:
            d += 1
            res += '/' + h_key[s:e]
            s = e
            e += 2
        res += '/' + h_key[s:]
        return '/' + res


def hash_key(key, use_hash=False):
    k = str(key)
    if use_hash:
        return sha3_256(k.encode("utf-8")).hexdigest()

    if len(k) < 10:
        k += '123456789'
    return encode64(k)


def cache_func(time: int, prefix=''):
    def dec(func):
        name = func.__name__
        file = func.__code__.co_filename

        def wrapper(*args, **kw):
            k = prefix + file + name
            for i in args:
                if type(i) in [str, int, float, list, dict, tuple, set]:
                    k = k + hash_key(i,True)

            has = Cache(use_hash=True).get(k, 'no_cached')
            if has != 'no_cached':
                return has

            f = func(*args, **kw)
            Cache(use_hash=True).put(k, f, time)
            return f

        return wrapper

    return dec


class Middleware:
    def __init__(self, request, middlewares=None) -> None:
        self.middlewares = [] if middlewares is None else middlewares
        self.request = request

    def next(self):
        for middleware in self.middlewares:
            handle=middleware(self.request)
            if handle.get('pass',False) is not True:
                return {'msg':handle.get('msg',None)}

        return True



import json

from renus.core.routing import Router
from renus.util.helper import hash_new_password


def run(args):
    if args[0] == 'setting':
        from app.extension.renus.setting.model import Setting
        Setting('').where({'_id': {'$ne': 1}}).delete(True)
        with open('extension/renus/setting/db.json','r') as db:
            d=json.loads(db.read())
            for item in d:
                Setting('').where({
                    'name':item['name']
                }).update(item, True)
        print('DB Settings Created.')

    if args[0] == 'translate':
        from app.extension.renus.translate.model import Translate
        Translate('').where({'_id':{'$ne':1}}).delete(True)
        with open('extension/renus/translate/db.json','r') as db:
            d=json.loads(db.read())
            for item in d:
                Translate('').where({
                    'key':item['key']
                }).update(item, True)
        print('DB Translates Created.')

    if args[0]=='super_admin':
        from app.extension.renus.role.model import Role
        from app.extension.renus.permission.model import Permission
        from app.extension.renus.user.model import User
        country_code = int(input('country_code: '))
        phone = int(input('phone: '))
        name = input('name: ')
        Permission('').where({
            'name': '*'
        }).update({}, True)
        permissions = Permission('').where({
            'name': '*'
        }).select('_id').get(False)
        res = []
        for p in permissions:
            res.append(p['_id'])
        Role('').where({
            'name': 'super_admin'
        }).update({'permission_ids': res,'active':True}, True)

        User('').where({
            'country_code': str(country_code),
            'phone': str(phone)
        }).update({
            'name': name,
            'username': 'admin',
            'password':hash_new_password('admin')
        }, True)
        User('').where({
            'country_code': str(country_code),
            'phone': str(phone)
        }).sync_roles(['super_admin'])
        print('Super Admin Created.\nPlease Change username and password.\nusername=admin\npassword=admin')


def import_by_string(name):
    return __import__(f"config.{name}", fromlist=[''])

class Config:
    lists = {}

    def __init__(self, name: str) -> None:
        if name in self.lists:
            self.config = self.lists[name]
        else:
            self.config = import_by_string(name)
            self.lists[name] = self.config

    def get(self, name:str, *default):
        return getattr(self.config,name,*default)



from renus.core.routing import Router
import routes.index
from app.extension.renus.permission.middleware import PermissionMiddleware
from app.extension.renus.permission.model import Permission

def run():
    all=Router().all
    names=set()
    names.add('delete_storage')
    for a in all:
        for method in all[a]:
            for route in all[a][method]:
                for md in route['middlewares']:
                    if isinstance(md,PermissionMiddleware):
                        print(route['path'], md.names)
                        for name in md.names:
                            names.add(name)

    for name in names:
        Permission('').where({
            'name':name
        }).update({},True)

    print('done')


import traceback
import json

from renus.core.serialize import jsonEncoder
from renus.core.status import Status
from renus.core.log import Log


def abort(msg, status_code: Status = Status.HTTP_400_BAD_REQUEST):
    if type(msg).__name__=='str':
        msg={'msg':msg}
    raise RuntimeError(msg, status_code)


def abort_if(condition: bool, msg, status_code: Status = Status.HTTP_400_BAD_REQUEST):
    if condition:
        if type(msg).__name__ == 'str':
            msg = {'msg': msg}
        raise RuntimeError(msg, status_code)


def abort_unless(condition: bool, msg, status_code: Status = Status.HTTP_400_BAD_REQUEST):
    if not condition:
        if type(msg).__name__ == 'str':
            msg = {'msg': msg}
        raise RuntimeError(msg, status_code)


def debug_response(exc: Exception,debug=False):
    if type(exc).__name__=='RuntimeError' and type(exc.args[0]).__name__=='dict' and len(exc.args)==2 and 'msg' in exc.args[0]:
        return exc.args[0],exc.args[1]

    status_code = Status.HTTP_500_INTERNAL_SERVER_ERROR
    traceback_obj = traceback.TracebackException.from_exception(
        exc, capture_locals=True
    )

    error = f"{traceback_obj.exc_type.__name__}: {str(traceback_obj)}"
    stacks = traceback_obj.stack.format()
    stack_list = []
    for stack in stacks:
        stack_list.append(stack.split('\n'))

    res = {
        'msg': error,
        'file': traceback.format_tb(exc.__traceback__),
        'stack': stack_list,
    }
    Log().error(json.dumps(
        res,
        ensure_ascii=False,
        allow_nan=True,
        indent=4,
        separators=(",", ":"),
        cls=jsonEncoder
    ))
    if not debug:
        return {
                   'msg': 'Internal Server Error',
               }, status_code
    traceback.print_tb(exc.__traceback__)
    return res, status_code



from setuptools import setup, find_packages

setup(
    name='renus',
    version='1.0.8',
    description='Renus Core Framework',
    url='https://github.com/smkoBa/renus',
    author='Smko Bayazidi',
    author_email='ba.smko@gmail.com',
    license='BSD',
    packages=find_packages(),
)


class bc:
    OKPINK = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def format(header,
                  left_format, cell_format, row_delim, col_delim):
    table = [[bc.OKPINK + 'command name' + bc.ENDC, bc.WARNING + 'description' + bc.ENDC]] + [
        [bc.OKGREEN + name + bc.ENDC, bc.OKCYAN + row + bc.ENDC] for name, row in header.items()]

    table_format = (len(header) + 1) * [[left_format] + len(header) * [cell_format]]

    col_widths = [max(
        len(format.format(cell, 0))
        for format, cell in zip(col_format, col))
        for col_format, col in zip(zip(*table_format), zip(*table))]
    return row_delim.join(
        col_delim.join(
            format.format(cell, width)
            for format, cell, width in zip(row_format, row, col_widths))
        for row_format, row in zip(table_format, table))


def Help():
    print(format({
        'app': 'Create New Package',
        'install renus/user/*': 'Install Apps',
        'backup': 'Backing up the entire project',
        'copy': 'Get a copy of the app to upload to the server',
        'permission': 'Add All Permissions to DB',
        'default super_admin': 'Create Super Admin'
    },
        '{:<{}}', '{:<{}}', '\n\n', ' | '))
